/**
 * @constructor Apemandocker
 * @param {object} config
 */
"use strict";

const execcli = require('execcli'),
    async = require('async'),
    os = require('os'),
    childProcess = require('child_process'),
    argx = require('argx'),
    stringcase = require('stringcase'),
    extend = require('extend');

/** @lends Apemandocker */
function Apemandocker(config) {
    let s = this;
    extend(s, config || {});
}

Apemandocker.prototype = {
    _docker: function (args, callback) {
        execcli('docker', args, {
            notfound: `install docker from https://docs.docker.com`
        }, callback);
    },
    /**
     * Build docker source.
     * @param {string} source - Name of source.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    buildImage: function (source, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['build', options, source], callback);
    },
    /**
     * Remove docker image.
     * @param {string} image - Name of image.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    removeImage: function (image, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['rmi', options, image], callback);
    },
    /**
     * List images.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    listImages: function (options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};

        async.waterfall([
            (callback) => {
                let command = `docker images`;
                childProcess.exec(command, callback);
            },
            (stdOut, stdErr, callback) => {
                if (stdErr) {
                    console.error(stdErr);
                }
                let lines = String(stdOut).trim().split(os.EOL);

                let headers = lines.shift().split(/\s\s+/g).map((key) => {
                    return stringcase.camelcase(String(key).trim().replace(/\s/g, '_'));
                });

                let results = lines.map((line) => {
                    let data = {};
                    line.split(/\s\s+/g).forEach((val, i) => {
                        let key = headers[i];
                        if (key) {
                            data[key] = val;
                        }
                    });
                    return data;
                });
                callback(null, results);
            }
        ], callback);
    },
    /**
     * Check if image exists.
     * @param {string} image - Name of image.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    imageExists: function (image, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s.listImages({}, (err, list) => {
            let exists = !err && list.filter((filtering)=> {
                    let repository = String(filtering.repository).trim(),
                        tag = String(filtering.tag).trim();
                    return (repository === image) || ([repository, tag].join(':') === image);
                }).length > 0;
            callback(exists);
        });
    },
    /**
     * Create a container.
     * @param {string} image - Name of image.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    createContainer: function (image, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['create', options, image], callback);
    },
    /**
     * Execute a command inside a container.
     * @param {string} container - Name of container.
     * @param {string} script - Script to execute.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    execInContainer: function (container, script, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['exec', options, container].concat(script), callback);
    },
    /**
     * Remove a container.
     * @param {string} container - Name of container.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    removeContainer: function (container, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['rm', options, container], callback);
    },
    /**
     * Start a container.
     * @param {string} container - Name of container.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    startContainer: function (container, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['start', options, container], callback);
    },
    /**
     * Run a container.
     * @param {string} image - Name of image.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    runContainer: function (image, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['run', options, image], callback);
    },
    /**
     * Stop a container.
     * @param {string} container - Name of container.
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    stopContainer: function (container, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['stop', options, container], callback);
    },
    /**
     * List containers
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    listContainers: function (options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};

        let all = !!options.all;

        const keys = [
            "ID",
            "Image",
            "Command",
            "CreatedAt",
            "RunningFor",
            "Ports",
            "Status",
            "Size",
            "Names",
            "Labels"
        ];
        let joiner = '|';
        async.waterfall([
            (callback) => {
                let format = keys.map((key) => {
                    return `{{.${key}}}`;
                }).join(joiner);
                let command = `docker ps${all ? ' -a' : ''} --format "${format}"`;
                childProcess.exec(command, callback);
            },
            (stdOut, stdErr, callback) => {
                if (stdErr) {
                    console.error(stdErr);
                }
                let results = String(stdOut).trim().split(os.EOL)
                    .filter((line) => {
                        return !!String(line).trim();
                    })
                    .map((line) => {
                        let data = {};
                        String(line).split(joiner).forEach((value, i) => {
                            let key = stringcase.camelcase(String(keys[i]).trim().replace(/\s/g, '_'))
                            data[key] = String(value).trim();
                        });
                        return data;
                    });
                callback(null, results);
            }
        ], callback);
    },
    /**
     * Check if container exists.
     * @param {string} container - Name of container.
     * @param options
     * @param {function} callback - Callback when done.
     */
    containerExists: function (container, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s.listContainers({
            all: true
        }, (err, list) => {
            let exists = !err && list.filter((filtering)=> {
                    return String(filtering.names).replace(/,/g, '') === container;
                }).length > 0;
            callback(exists);
        });
    },
    /**
     * Detect a container running.
     * @param {string} container - Name of container.
     * @param {function} callback - Callback when done.
     */
    isContainerRunning: function (container, callback) {
        let command = `docker inspect -f {{.State.Running}} ${container}`;
        childProcess.exec(command, (err, stdOut, stdErr) => {
            let isRunning = !err && (String(stdOut).trim().toLowerCase() === "true");
            callback(isRunning);
        });
    },
    /**
     * Log container.
     * @param container
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    logContainer: function (container, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['logs', options, container], callback);
    },
    /**
     * Check if network exists.
     * @param network
     * @param {object} [options] - Optional settings.
     * @param {function} callback - Callback when done.
     */
    networkExists: function (network, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s.listNetworks(options, (err, networks)=> {
            if (!err) {
                for (let i = 0, len = networks.length; i < len; i++) {
                    let hit = networks[i].name === network;
                    if (hit) {
                        callback(true);
                        return;
                    }
                }
            }
            callback(false);
        });
    },
    /**
     * Open terminal.
     * @param {string} container - Name of container.
     * @param {object} [options] - Optional settings.
     * @param {string} [options.shell='bash'] - Shell for terminal
     */
    terminal: function (container, options) {
        let s = this;
        let args = argx(arguments);
        options = args.pop('object') || {};
        return childProcess.spawn('docker',
            ['exec', '-it', container, options.shell || 'bash'],
            {
                stdio: 'inherit'
            }
        );
    },
    createNetwork: function (name, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['network', 'create', options, name], callback);
    },
    listNetworks: function (options, callback) {
        let s = this;
        let command = `docker network ls`;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        childProcess.exec(command, (err, stdOut, stdErr) => {
            if (stdErr) {
                console.error(stdErr);
            }
            if (err) {
                callback(err);
            } else {
                let lines = String(stdOut).trim().split(os.EOL);

                let headers = lines.shift().split(/\s\s+/g).map((key) => {
                    return stringcase.camelcase(String(key).trim().replace(/\s/g, '_'));
                });

                let results = lines.map((line) => {
                    let data = {};
                    line.split(/\s\s+/g).forEach((val, i) => {
                        let key = headers[i];
                        if (key) {
                            data[key] = val;
                        }
                    });
                    return data;
                });
                callback(null, results);
            }

        });
    },
    removeNetwork: function (name, options, callback) {
        let s = this;
        let args = argx(arguments);
        callback = args.pop('function') || argx.noop;
        options = args.pop('object') || {};
        s._docker(['network', 'rm', options, name], callback);
    }


};

module.exports = Apemandocker;
